@@ -1,7 +1,31 @@
-//  This work is based on the original netsed 0.01c by Michal Zalewski.
-//  Please contact Julien VdG <julien@silicone.homelinux.org> if you encounter
-//  any problems with this version.
-//  The changes compared to version 0.01c are related in the NEWS file.
+/*
+  netsed 1.1       (C) 2010-2012  Julien VdG <julien@silicone.homelinux.org>
+  --------------------------------------------------------------------------
+
+  This work is based on the original netsed:
+      netsed 0.01c      (C) 2002  Michal Zalewski <lcamtuf@ids.pl>
+
+  Please contact Julien VdG <julien@silicone.homelinux.org> if you encounter
+  any problems with this version.
+  The changes compared to version 0.01c are related in the NEWS file.
+
+
+This program is free software; you can redistribute it and/or
+modify it under the terms of the GNU General Public License
+as published by the Free Software Foundation; either version 2
+of the License, or (at your option) any later version.
+
+This program is distributed in the hope that it will be useful,
+but WITHOUT ANY WARRANTY; without even the implied warranty of
+MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
+GNU General Public License for more details.
+
+You should have received a copy of the GNU General Public License
+along with this program; if not, write to the Free Software
+Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
+
+*/
+
 
 ///@mainpage
 ///
@@ -64,6 +88,7 @@
 /// The TODO file:
 ///@verbinclude TODO
 
+
 #include <stdio.h>
 #include <unistd.h>
 #include <sys/types.h>
@@ -82,18 +107,26 @@
 #include <netdb.h>
 #include <time.h>
 
+#ifdef __linux__
 /// Define for transparent proxy with linux netfilter.
 /// Else use getsockname() supposing the socket receive the original
 /// destination information directly.
 #define LINUX_NETFILTER
+#endif
 
 #ifdef LINUX_NETFILTER
 #include <limits.h>
 #include <linux/netfilter_ipv4.h>
 #endif
 
+/// Define to use getopt_long: GNU extension, should check _GNU_SOURCE
+#define PARSE_LONG_OPT
+#ifdef PARSE_LONG_OPT
+#include <getopt.h>
+#endif
+
 /// Current version (recovered by Makefile for several release checks)
-#define VERSION "1.00b"
+#define VERSION "1.1"
 /// max size for buffers
 #define MAX_BUF  100000
 
@@ -166,8 +199,26 @@ struct tracker_s {
 
 /// Store current time (just after select returned).
 time_t now;
+
 /// Listening socket.
 int lsock;
+
+// Command line parameters are parsed to the following global variables.
+
+/// Address family used for parameter resolution
+int family = AF_UNSPEC;
+
+/// TCP or UDP.
+int tcp;
+
+/// Local Port.
+char* lport;
+
+/// Remote Host.
+char* rhost;
+/// Remote Port.
+char* rport;
+
 /// Number of rules.
 int rules;
 /// Array of all rules.
@@ -182,11 +233,31 @@ struct tracker_s * connections = NULL;
 /// True when SIGINT signal was received.
 volatile int stop=0;
 
+/// Display an error message followed by short usage information.
+/// @param why the error message.
+void short_usage_hints(const char* why) {
+  if (why) ERR("Error: %s\n\n",why);
+  ERR("Usage: netsed [option] proto lport rhost rport rule1 [ rule2 ... ]\n\n");
+  ERR("  use netsed -h for more information on usage.\n");
+  exit(1);
+}
+
+
 /// Display an error message followed by usage information.
 /// @param why the error message.
 void usage_hints(const char* why) {
-  ERR("Error: %s\n\n",why);
-  ERR("Usage: netsed proto lport rhost rport rule1 [ rule2 ... ]\n\n");
+  if (why) ERR("Error: %s\n\n",why);
+  ERR("Usage: netsed [option] proto lport rhost rport rule1 [ rule2 ... ]\n\n");
+#ifdef PARSE_LONG_OPT
+  ERR("  options - can be --ipv4 or -4 to force address resolution in IPv4,\n");
+  ERR("            --ipv6 or -6 to force address resolution in IPv6,\n");
+  ERR("            --ipany to resolve the address in either IPv4 or IPv6.\n");
+  ERR("          - --help or -h to display this usage informations.\n");
+#else
+  ERR("  options - can be nothing, -4 to force address resolution in IPv4\n");
+  ERR("            or -6 to force address resolution in IPv6.\n");
+  ERR("          - -h to display this usage informations.\n");
+#endif
   ERR("  proto   - protocol specification (tcp or udp)\n");
   ERR("  lport   - local port to listen on (see README for transparent\n");
   ERR("            traffic intercepting on some systems)\n");
@@ -211,6 +282,7 @@ void usage_hints(const char* why) {
   exit(1);
 }
 
+
 /// Helper function to free a tracker_s item.
 /// csa will be freed if needed, sockets will be closed
 /// @param conn pointer to free.
@@ -367,6 +439,86 @@ void shrink_to_binary(struct rule_s* r) {
   }
 }
 
+/// parse the command line parameters
+/// @param argc number of arguments
+/// @param argv array of string parameters
+void parse_params(int argc,char* argv[]) {
+  int i;
+
+  // parse options, GNU allows us to use long options
+#ifdef PARSE_LONG_OPT
+  static struct option long_options[] = {
+    {"ipv4", 0, 0, '4'},
+    {"ipv6", 0, 0, '6'},
+    {"help", 0, 0, 'h'},
+    {"ipany", 0, &family, AF_UNSPEC},
+    {0, 0, 0, 0}
+  };
+
+  while ((i = getopt_long(argc, argv, "46h", long_options, NULL)) != -1)
+#else
+  while ((i = getopt(argc, argv, "46h")) != -1)
+#endif
+  {
+    switch(i) {
+    case 0: // long option
+      break;
+    case '4':
+      family = AF_INET;
+      break;
+    case '6':
+      family = AF_INET6;
+      break;
+    case 'h':
+      usage_hints(NULL);
+    default:
+      usage_hints("unsupported optional parameter");
+    }
+  }
+
+  // parse remaining positional parameters
+  if (argc<optind+5) short_usage_hints("not enough parameters");
+
+  // protocole
+  if (strcasecmp(argv[optind],"tcp")*strcasecmp(argv[optind],"udp")) short_usage_hints("incorrect protocol");
+  tcp = strncasecmp(argv[optind++], "udp", 3);
+
+  // local port
+  lport = argv[optind++];
+
+  // remote host & port
+  rhost = argv[optind++];
+  rport = argv[optind++];
+
+  // allocate rule arrays, rule number is number of params after 5
+  rule=malloc((argc-optind)*sizeof(struct rule_s));
+  rule_live=malloc((argc-optind)*sizeof(int));
+  // parse rules
+  for (i=optind;i<argc;i++) {
+    char *fs=0, *ts=0, *cs=0;
+    printf("[*] Parsing rule %s...\n",argv[i]);
+    fs=strchr(argv[i],'/');
+    if (!fs) error("missing first '/' in rule");
+    fs++;
+    ts=strchr(fs,'/');
+    if (!ts) error("missing second '/' in rule");
+    *ts=0;
+    ts++;
+    cs=strchr(ts,'/');
+    if (cs) { *cs=0; cs++; }
+    rule[rules].forig=fs;
+    rule[rules].torig=ts;
+    if (cs && *cs) /* Only non-trivial quantifiers count. */
+      rule_live[rules]=atoi(cs); else rule_live[rules]=-1;
+    shrink_to_binary(&rule[rules]);
+//    printf("DEBUG: (%s) (%s)\n",rule[rules].from,rule[rules].to);
+    rules++;
+  }
+
+  printf("[+] Loaded %d rule%s...\n", rules, (rules > 1) ? "s" : "");
+
+}
+
 /// Bind and optionally listen to a socket for netsed server port.
 /// @param af      address family.
 /// @param tcp     1 tcp, 0 udp.
@@ -393,6 +545,11 @@ void bind_and_listen(int af, int tcp, const char *portstr) {
       continue;
     setsockopt(lsock, SOL_SOCKET, SO_REUSEADDR, &one, sizeof(one));
     //fcntl(lsock,F_SETFL,O_NONBLOCK);
+    /* Make our best to decide on dual-stacked listener. */
+    one = (family == AF_UNSPEC) ? 0 /* All families */ : 1; /* Preconditioned addr */
+    if (res->ai_family == AF_INET6)
+      if (setsockopt(lsock, IPPROTO_IPV6, IPV6_V6ONLY, &one, sizeof(one)))
+        printf("    Failed to unset IPV6_V6ONLY: %s.\n", strerror(errno));
     if (bind(lsock, res->ai_addr, res->ai_addrlen) < 0) {
       ERR("bind(): %s", strerror(errno));
       close(lsock);
@@ -530,52 +687,25 @@ void sig_int(int signo)
 
 /// This is main...
 int main(int argc,char* argv[]) {
-  int i, ret;
+  int ret;
   in_port_t fixedport = 0;
   struct sockaddr_storage fixedhost;
   struct addrinfo hints, *res, *reslist;
-  int tcp;
   struct tracker_s * conn;
 
   memset(&fixedhost, '\0', sizeof(fixedhost));
   printf("netsed " VERSION " by Julien VdG <julien@silicone.homelinux.org>\n"
          "      based on 0.01c from Michal Zalewski <lcamtuf@ids.pl>\n");
   setbuffer(stdout,NULL,0);
-  if (argc<6) usage_hints("not enough parameters");
-  if (strcasecmp(argv[1],"tcp")*strcasecmp(argv[1],"udp")) usage_hints("incorrect protocol");
-  tcp = strncasecmp(argv[1], "udp", 3);
-  // allocate rule arrays, rule number is number of params after 5
-  rule=malloc((argc-5)*sizeof(struct rule_s));
-  rule_live=malloc((argc-5)*sizeof(int));
-  // parse rules
-  for (i=5;i<argc;i++) {
-    char *fs=0, *ts=0, *cs=0;
-    printf("[*] Parsing rule %s...\n",argv[i]);
-    fs=strchr(argv[i],'/');
-    if (!fs) error("missing first '/' in rule");
-    fs++;
-    ts=strchr(fs,'/');
-    if (!ts) error("missing second '/' in rule");
-    *ts=0;
-    ts++;
-    cs=strchr(ts,'/');
-    if (cs) { *cs=0; cs++; }
-    rule[rules].forig=fs;
-    rule[rules].torig=ts;
-    if (cs) rule_live[rules]=atoi(cs); else rule_live[rules]=-1;
-    shrink_to_binary(&rule[rules]);
-//    printf("DEBUG: (%s) (%s)\n",rule[rules].from,rule[rules].to);
-    rules++;
-  }
 
-  printf("[+] Loaded %d rule%s...\n", rules, (rules > 1) ? "s" : "");
+  parse_params(argc, argv);
 
   memset(&hints, '\0', sizeof(hints));
-  hints.ai_family = AF_UNSPEC;
+  hints.ai_family = family;
   hints.ai_flags = AI_CANONNAME;
   hints.ai_socktype = tcp ? SOCK_STREAM : SOCK_DGRAM;
 
-  if ((ret = getaddrinfo(argv[3], argv[4], &hints, &reslist))) {
+  if ((ret = getaddrinfo(rhost, rport, &hints, &reslist))) {
     ERR("getaddrinfo(): %s\n", gai_strerror(ret));
     error("Impossible to resolve remote address or port.");
   }
@@ -598,17 +728,17 @@ int main(int argc,char* argv[]) {
     error("Failed in resolving remote host.");
 
   if (fixedhost.ss_family && fixedport)
-    printf("[+] Using fixed forwarding to %s,%s.\n",argv[3],argv[4]);
+    printf("[+] Using fixed forwarding to %s,%s.\n",rhost,rport);
   else if (fixedport)
-    printf("[+] Using dynamic (transparent proxy) forwarding with fixed port %s.\n",argv[4]);
+    printf("[+] Using dynamic (transparent proxy) forwarding with fixed port %s.\n",rport);
   else if (fixedhost.ss_family)
-    printf("[+] Using dynamic (transparent proxy) forwarding with fixed addr %s.\n",argv[3]);
+    printf("[+] Using dynamic (transparent proxy) forwarding with fixed addr %s.\n",rhost);
   else
     printf("[+] Using dynamic (transparent proxy) forwarding.\n");
 
-  bind_and_listen(fixedhost.ss_family, tcp, argv[2]);
+  bind_and_listen(fixedhost.ss_family, tcp, lport);
 
-  printf("[+] Listening on port %s/%s.\n", argv[2], argv[1]);
+  printf("[+] Listening on port %s/%s.\n", lport, (tcp)?"tcp":"udp");
 
   signal(SIGPIPE, SIG_IGN);
   struct sigaction sa;
